# -*- coding: utf-8 -*-
from __future__ import annotations

from dataclasses import dataclass, field
from io import TextIOWrapper
from typing import List, Tuple, Literal, Iterable, Union, Optional, TextIO, TYPE_CHECKING

if TYPE_CHECKING:
    from pybiotk.io.gtf import GTF


@dataclass
class TransInfo:
    transcript_id: str = field(default=None)
    transcript_name: str = field(default=None)
    transcript_type: str = field(default=None)
    gene_id: str = field(default=None)
    gene_type: str = field(default=None)
    gene_name: str = field(default=None)
    strand: Literal['+', '-'] = field(default="+")

    @classmethod
    def init_by_gtf(cls, gtf: GTF, attr: Optional[str] = None):
        return cls(attr, gtf.transcript_name(), gtf.transcript_type(),
                   gtf.gene_id(), gtf.gene_name(), gtf.gene_type(), gtf.strand)

    def __str__(self):
        return "\t".join(str(s) for s in list(self.__dict__.values()))


@dataclass
class GeneInfo:
    gene_id: str = field(default=None)
    gene_name: str = field(default=None)
    gene_type: str = field(default=None)
    strand: Literal['+', '-'] = field(default='+')

    @classmethod
    def init_by_gtf(cls, gtf: GTF, attr: str):
        return cls(attr, gtf.gene_name(), gtf.gene_type(), gtf.strand)

    def __str__(self):
        return "\t".join(str(s) for s in list(self.__dict__.values()))


@dataclass
class Bed6:
    chrom: str = field(default=None)
    start: int = field(default=0)
    end: int = field(default=0)
    name: str = field(default=None)
    score: str = field(default=None)
    strand: Literal['+', '-'] = field(default='+')

    def __post_init__(self):
        self.start = int(self.start)
        self.end = int(self.end)

    @classmethod
    def init_by_gtf(cls, gtf: GTF, attr: str):
        return cls(gtf.chrom, gtf.start-1, gtf.end, attr, gtf.score, gtf.strand)

    def __str__(self):
        return "\t".join(str(s) for s in list(self.__dict__.values()))


@dataclass
class Bed12:
    chrom: str = field(default=None)
    start: int = field(default=0)
    end: int = field(default=0)
    name: str = field(default=None)
    score: int = field(default=0, repr=False)
    strand: Literal['+', '-'] = field(default='+')
    thickStart: int = field(default=0, repr=False)
    thickEnd: int = field(default=0, repr=False)
    itemRgb: int = field(default=0, repr=False)
    blockCount: int = field(default=1, repr=False)
    blockSizes: str = field(default=None, repr=False)
    blockStarts: str = field(default=None, repr=False)

    def __post_init__(self):
        self.start = int(self.start)
        self.end = int(self.end)
        self.thickStart = int(self.thickStart)
        self.thickEnd = int(self.thickEnd)

    @classmethod
    def init_by_gtf(cls, gtf: GTF, attr: str, thickStart: Optional[int] = 0, thickEnd: Optional[int] = 0):
        chrom = gtf.chrom
        start = gtf.start - 1
        end = gtf.end
        name = attr
        score = 0 if gtf.score == '.' else int(gtf.score)
        strand = gtf.strand
        thickStart = thickStart
        thickEnd = thickEnd
        itemRgb = 0
        blockCount = 1
        blockSizes = "{size},".format(size=gtf.end - gtf.start + 1)
        blockStarts = "{start},".format(start=0)
        return cls(chrom, start, end, name, score, strand, thickStart,
                   thickEnd, itemRgb, blockCount, blockSizes, blockStarts)

    def __str__(self):
        return "\t".join(str(s) for s in list(self.__dict__.values()))

    def update(self, gtf: GTF):
        self.end = gtf.end
        self.blockSizes += "{size},".format(size=gtf.end - gtf.start + 1)
        self.blockStarts += "{start},".format(start=gtf.start - 1 - self.start)
        self.blockCount += 1

    def exons(self) -> List[Tuple[int, int]]:
        exons = sorted((self.start+int(y), self.start+int(y)+int(x)) for x, y in zip(
            self.blockSizes.rstrip(',').split(','), self.blockStarts.rstrip(',').split(',')))
        return exons


@dataclass
class Intron:
    chrom: str = field(default=None)
    start: int = field(default=0)
    end: int = field(default=0)
    name: str = field(default=None)
    score: str = field(default='.')
    strand: Literal['+', '-'] = field(default='+')

    def __post_init__(self):
        self.start = int(self.start)
        self.end = int(self.end)

    @classmethod
    def init_by_gtf(cls, exon_1: GTF, exon_2: GTF, attr: str):
        return cls(exon_1.chrom, exon_1.end, exon_2.start-1, attr, '.', exon_1.strand)

    def __str__(self):
        return "\t".join(str(s) for s in list(self.__dict__.values()))


class Openbed:
    def __init__(self, filepath_or_buffer: Union[str, TextIO], mode: Literal["bed6", "bed12"] = "bed6"):
        if isinstance(filepath_or_buffer, TextIOWrapper):
            self.filename: str = filepath_or_buffer.name
            self.bed: TextIO = filepath_or_buffer
        else:
            self.filename: str = filepath_or_buffer
            self.bed: TextIO = open(filepath_or_buffer)
        self.mode: str = mode

    def __repr__(self):
        return f"Openbed(filename='{self.filename}', mode='{self.mode}')"

    __str__ = __repr__

    @staticmethod
    def str_to_bed(line: str, mode: Literal["bed6", "bed12"]) -> Union[Bed6, Bed12]:
        args = line.rstrip("\r\n").split("\t")
        args_len = len(args)
        if mode == "bed6":
            if args_len < 6:
                args.extend((6-len(args))*[''])
            if args_len > 6:
                args = args[:6]
            bed = Bed6(*args)
        elif mode == "bed12":
            if args_len < 12:
                args.extend((12-len(args))*[''])
            if args_len > 12:
                args = args[:12]
            bed = Bed12(*args)
        else:
            raise RuntimeError(f"{mode} not support.")
        return bed

    def close(self):
        self.bed.close()

    def __iter__(self) -> Iterable[Union[Bed6, Bed12]]:
        for line in self.bed:
            yield self.str_to_bed(line, self.mode)

    def __next__(self) -> Union[Bed6, Bed12]:
        return self.str_to_bed(next(self.bed), self.mode)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, traceback):
        self.close()



#!/usr/bin/env python3
import sys
from abc import ABC, abstractmethod
from typing import Dict, List, Literal

import numpy as np
import pyBigWig
from numpy.lib.stride_tricks import sliding_window_view


class TrackFile(ABC):

    @property
    @abstractmethod
    def dchroms(self) -> Dict[str, int]: ...

    @abstractmethod
    def close(self): ...

    @abstractmethod
    def stats(self, chrom: str, start: int, end: int, stat: Literal['mean', 'max', 'min', 'sum', 'coverage', 'std'] = 'mean') -> np.ndarray:
        """
        assert turn None and np.nan to 0
        Args:
            chrom (str, optional): Defaults to None.
            start (int, optional): Defaults to None.
            end (int, optional):   Defaults to None.
            stat (Literal[, optional):  Defaults to 'mean'.

        Returns:
            np.ndarray: shape: (n,)
        """
        pass

    @abstractmethod
    def values(self, chrom: str, start: int, end: int) -> np.ndarray:
        """
        assert turn None and np.nan to 0
        Returns:
            np.ndarray: [, (n,end-start)]
        """
        pass

    @abstractmethod
    def __len__(self) -> int: ...

    def coverage(self, chrom: str, start: int, end: int) -> np.ndarray:
        """
        Returns:
            np.ndarray: shape: [-1, ]
        """
        a = self.values(chrom, start, end)
        return np.count_nonzero(a, axis=a.ndim-1) / a.shape[-1]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        if any([exc_type, exc_val, exc_tb]):
            return False


class Openbwn(TrackFile):
    def __init__(self, files: List[str]):
        self.files = files
        self.bws = [pyBigWig.open(x) for x in files]
        self.headers: List[Dict[str, int]] = [x.header() for x in self.bws]
        self.chroms: List[Dict[str, int]] = [x.chroms() for x in self.bws]

    @property
    def dchroms(self) -> Dict[str, int]:
        return self.chroms[0]

    def stats(self, chrom: str, start: int, end: int, stat: Literal['mean', 'max', 'min', 'sum', 'coverage', 'std'] = 'mean') -> np.ndarray:
        try:
            a = np.nan_to_num(np.array([bw.stats(chrom, start, end, exact=True, type=stat)[0] for bw in self.bws]))
            a[a is None] = 0
            return np.abs(a.astype(np.float32))
        except RuntimeError:
            sys.stderr.write(f'Invalid interval bounds! {chrom}:{start}-{end}')
            raise

    def values(self, chrom: str, start: int = None, end: int = None) -> np.ndarray:
        try:
            a = np.nan_to_num(np.array([bw.values(chrom, start, end) for bw in self.bws]))
            a[a is None] = 0
            return np.abs(a.astype(np.float32))
        except RuntimeError:
            sys.stderr.write(f'Invalid interval bounds! {chrom}:{start}-{end}')
            raise

    @staticmethod
    def coverage_bins_np_values(a: np.ndarray, nbins: int):
        assert a.shape[-1] % nbins == 0, 'length should be divisible by nbins '
        try:
            a2 = a.reshape((-1, int(a.shape[-1] / nbins)))
            return (np.count_nonzero(a2, axis=1) / a2.shape[-1]).reshape((a.shape[0], -1))
        except ValueError:
            sys.stderr.write(f"length of array is not divisible by nbins: {nbins}\n")
            raise

    def coverage_bins(self, chrom: str, start: int, end: int, nbins: int) -> np.ndarray:
        a = self.values(chrom, start, end)
        return self.coverage_bins_np_values(a, nbins)

    @staticmethod
    def scale_region_values_np_values(a: np.ndarray, length: int = 1000, nbins: int = 10) -> np.ndarray:
        """
        Returns:
            np.ndarray: shape: [-1, length/nbins]
        """
        n_base = length // nbins
        window_len = a.shape[-1] // n_base
        if window_len == 0:
            return np.zeros((a.shape[0], n_base))
        a = a[:, :window_len*n_base]
        return a.reshape(-1, window_len).mean(axis=1).reshape(a.shape[0], -1)

    def scale_region_values(self, chrom: str, start: int, end: int, length: int = 1000, nbins: int = 10):
        a = self.values(chrom, start, end)
        return self.scale_region_values_np_values(a, length, nbins)

    @staticmethod
    def coverage_sliding_window_np_values(a: np.ndarray, nbins: int) -> np.ndarray:
        a2 = sliding_window_view(a, (a.shape[0], nbins))
        a3 = (a2 > 0).sum(axis=3) / nbins
        return a3.squeeze(axis=0).transpose()

    def coverage_sliding_window(self, chrom: str, start: int, end: int, nbins: int):
        a = self.values(chrom, start, end)
        return self.coverage_sliding_window_np_values(a, nbins)

    def close(self):
        [x.close() for x in self.bws]

    def __len__(self) -> int:
        return len(self.bws)


class Openbw(TrackFile):
    def __init__(self, file: str):
        self.file = file
        self.bw = pyBigWig.open(file)
        self.header: Dict[str, int] = self.bw.header()
        self.chroms: Dict[str, int] = self.bw.chroms()

    @property
    def dchroms(self) -> Dict[str, int]:
        return self.chroms

    def stats(self, chrom: str, start: int, end: int, stat: Literal['mean', 'max', 'min', 'sum', 'coverage', 'std'] = 'mean') -> np.ndarray:
        try:
            a = np.nan_to_num(np.array(self.bw.stats(chrom, start, end, exact=True, type=stat)[0]))
            return np.abs(a.astype(np.float32))
        except RuntimeError:
            sys.stderr.write(f'Invalid interval bounds! {chrom}:{start}-{end}')
            raise

    def values(self, chrom: str, start: int = None, end: int = None) -> np.ndarray:
        try:
            a = np.nan_to_num(np.array(self.bw.values(chrom, start, end)))
            a[a is None] = 0
            return np.abs(a.astype(np.float32))
        except RuntimeError:
            sys.stderr.write(f'Invalid interval bounds! {chrom}:{start}-{end}')
            raise

    @staticmethod
    def coverage_bins_np_values(a: np.ndarray, nbins: int):
        assert a.shape[-1] % nbins == 0, 'length should be divisible by nbins '
        try:
            a2 = a.reshape((-1, int(a.shape[-1] / nbins)))
            return np.count_nonzero(a2, axis=1) / a2.shape[-1]
        except ValueError:
            sys.stderr.write(f"length of array is not divisible by nbins: {nbins}\n")
            raise

    def coverage_bins(self, chrom: str, start: int, end: int, nbins: int) -> np.ndarray:
        a = self.values(chrom, start, end)
        return self.coverage_bins_np_values(a, nbins)

    @staticmethod
    def scale_region_values_np_values(a: np.ndarray, length: int = 1000, nbins: int = 10) -> np.ndarray:
        """
        Returns:
            np.ndarray: shape: [-1, length/nbins]
        """
        n_base = length // nbins
        window_len = a.shape[-1] // n_base
        if window_len == 0:
            return np.zeros(n_base)
        a = a[:window_len*n_base]
        return a.reshape(-1, window_len).mean(axis=1)

    def scale_region_values(self, chrom: str, start: int, end: int, length: int = 1000, nbins: int = 10):
        a = self.values(chrom, start, end)
        return self.scale_region_values_np_values(a, length, nbins)

    @staticmethod
    def coverage_sliding_window_np_values(a: np.ndarray, nbins: int) -> np.ndarray:
        a2 = sliding_window_view(a, nbins)
        return (a2 > 0).sum(axis=1) / nbins

    def coverage_sliding_window(self, chrom: str, start: int, end: int, nbins: int):
        a = self.values(chrom, start, end)
        return self.coverage_sliding_window_np_values(a, nbins)

    def close(self):
        self.bw.close()

    def __len__(self) -> int:
        return 1



#!/usr/bin/env python3
"""
Remove duplicated rows of the table by specified columns and keep the contents of other columns.
Author: liqiming@whu.edu.cn
"""
import argparse
import sys

import pandas as pd

from pybiotk.utils import read_table, write_table, ignore


def df_merge_row(df, by=None, columns=None, method=("distinct",), sep=",", count=False):
    if not columns:
        return df.drop_duplicates(subset=by, keep='first')

    def generate_content(x):
        data = {}
        for c, m in zip(columns, method):
            try:
                series = pd.to_numeric(x[c])
                if m == "mean":
                    data[c] = series.mean()
                elif m == 'medium':
                    data[c] = series.median()
            except ValueError:
                series = x[c]
            if m == 'distinct':
                data[c] = sep.join(map(str, series.unique()))
            elif m == 'collapse':
                data[c] = sep.join(map(str, series))
            elif m == 'max':
                data[c] = series.max()
            elif m == 'min':
                data[c] = series.min()
            elif m == 'sum':
                data[c] = series.sum()
            elif m == 'first':
                data[c] = series.to_list()[0]
            elif m == 'last':
                data[c] = series.to_list()[-1]
            elif m == 'count':
                data[c] = series.count()
            elif m == 'count_distinct':
                data[c] = pd.Series(series.unique()).count()
            else:
                if c not in data:
                    data[c] = sep.join(map(str, series.unique()))
        if count:
            data["count"] = len(x[columns[0]])
        return pd.Series(data)

    merge_df = df.groupby(by).apply(generate_content).reset_index()
    return merge_df


def main(filename, output, header, noheader, by, columns=None, method="distinct", sep=None, delim=",", count=False):
    if noheader or header < 0:
        header = None
    df = read_table(filename, header=header, sep=sep)
    out_header = True if header is not None else False
    if df.empty:
        write_table(df, output, header=out_header)
        return

    by = by.split(",")
    method = method.split(",")

    if columns:
        by_name = [df.columns[int(i)] for i in by]
        columns = columns.split(",")
        if len(method) == 1:
            method = len(columns) * method
        assert len(columns) == len(method)
        try:
            column_name = [df.columns[int(i)] for i in columns if i not in by]
            output_name = [df.columns[int(i)] for i in sorted(by + columns)]
            if count:
                output_name.append("count")
        except RuntimeError:
            column_name = None
            output_name = by_name
    else:
        by_name = [df.columns[int(i)] for i in by]
        column_name = [i for i in df.columns if i not in by_name]
        if len(method) == 1:
            method = len(column_name) * method
        assert len(column_name) == len(method)
        output_name = df.columns.to_list()
        if count:
            output_name.append("count")

    out_df = df_merge_row(df, by_name, column_name, method, delim, count)
    out_df = out_df[output_name]
    write_table(out_df, output, index=False, header=out_header)


@ignore
def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-i", "--input", dest="input", type=argparse.FileType('r'),
                        default=(None if sys.stdin.isatty() else sys.stdin), help="Input table file.")
    parser.add_argument("-o", "--output", dest="output", type=argparse.FileType('w'),
                        default=sys.stdout, help="Output file, discard this parameter to print output to stdout.")
    parser.add_argument("--header", dest="header", type=int, default=0,
                        help="The selected row will be the header, 0 based.")
    parser.add_argument("--noheader", dest="noheader", action="store_true",
                        help="If no header, this option overrides --header.")
    parser.add_argument("-by", dest="by", type=str, default="0",
                        help="The same values in the selected columns will be merged, 0 based, use ',' to separate.")
    parser.add_argument("-c", "--columns", dest="columns", type=str, default=None,
                        help=("Other columns selected will be shown in the output, "
                              "discard this parameter to show all columns, 0 based, use ',' to separate. "
                              "If an illegal parameter is set, no other columns will be appended to the output table."))
    parser.add_argument("-m", "--method", dest="method", type=str, default="distinct",
                        help=("Specify the operation that should be applied to -c. "
                              "Valid operations: sum, min, max, mean, median, collapse, distinct, first, last, count, count_distinct. "
                              "use ',' to separate."))
    parser.add_argument("-sep", dest="sep", type=str, default=None,
                        help=("Specify a custom delimiter to separate the input file,"
                              "discard this parameter, the separator will be automatically determined according to the suffix"))
    parser.add_argument("-d", "--delimiter", dest="delim", type=str, default=",",
                        help="Specify a custom delimiter for the collapse operations.")
    parser.add_argument("--count", dest="count", action="store_true",
                        help="Add a count column to the output table.")
    args = parser.parse_args()

    if args.input is None:
        parser.parse_args(['-h'])

    main(**args.__dict__)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
import argparse
import time
from typing import Sequence

from pybiotk.io import FastqFile, FastqPair, OpenFqGzip
from pybiotk.utils import logging


def single_end(input_fq: str, output: str, min_len: int = 15):
    logging.info("Single end mode...")
    too_short_reads = 0
    output_reads = 0
    with FastqFile(input_fq) as fqi, OpenFqGzip(output) as fqo:
        for fq in fqi.uniq():
            if len(fq.sequence) < min_len:
                too_short_reads += 1
                continue
            fqo.write_fastx_record(fq)
            output_reads += 1
        input_reads = fqi.ptr
        fqi.ptr = None
    duplicate_ratio = (input_reads - output_reads) * 100 / input_reads
    logging.info((f"result summary:\nread pairs too short (<{min_len}nt): {too_short_reads}\ninput read pairs: {input_reads}\n"
                  f"output read pairs: {output_reads}\nduplicate ratio: {duplicate_ratio:.2f}%"))


def pair_end(input_r1: str, input_r2: str, output_r1: str, output_r2: str, min_len: int = 15):
    logging.info("Pair end mode...")
    too_short_reads = 0
    output_reads = 0

    with FastqPair(input_r1, input_r2) as pairfq, OpenFqGzip(output_r1) as r1, OpenFqGzip(output_r2) as r2:
        for fq1, fq2 in pairfq.uniq():
            if min(len(fq1.sequence), len(fq2.sequence)) < min_len:
                too_short_reads += 1
                continue
            r1.write_fastx_record(fq1)
            r2.write_fastx_record(fq2)
            output_reads += 1
        input_reads = pairfq.ptr
        pairfq.ptr = None
    duplicate_ratio = (input_reads - output_reads) * 100 / input_reads
    logging.info((f"result summary:\nread pairs too short (<{min_len}nt): {too_short_reads}\ninput read pairs: {input_reads}\n"
                  f"output read pairs: {output_reads}\nduplicate ratio: {duplicate_ratio:.2f}%"))


def main(input_files: Sequence[str], output_files: Sequence[str], min_len: int = 15):
    input_str = " ".join(input_files)
    output_str = " ".join(output_files)
    assert len(input_files) == len(output_files)
    logging.info(f"Processing {input_str} ...")

    start = time.perf_counter()
    if len(input_files) == 1:
        single_end(input_files[0], output_files[0], min_len)
    else:
        pair_end(*input_files[:2], *output_files[:2], min_len=min_len)
    end = time.perf_counter()
    logging.info(f"task finished in {end-start:.2f}s, output saved in {output_str}.")


def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-i", dest="input", type=str, nargs="+", required=True, help="input *.fastq.gz files.")
    parser.add_argument("-o", dest="output", type=str, nargs="+", required=True, help="output *.fastq.gz files.")
    parser.add_argument("-m", dest="min_len", type=int, default=15, help="min length.")
    args = parser.parse_args()
    main(args.input, args.output, args.min_len)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
import argparse
import os
from collections import defaultdict

import pysam


def main(originfa, path, outdir, more_x="C"):
    if not os.path.exists(path):
        os.makedirs(path)
    reads_dict = defaultdict(list)
    a = []
    b = []
    with open(path) as reads:
        for line in reads:
            fileds = line.split()
            reads_dict[fileds[1]].append(fileds[0])
            a.append(fileds[0])
            b.append(fileds[1])
    a = set(a)
    b = set(b)

    newset = b - a
    print(f"contained only by other sequences: {len(a - b)}")
    print(f"is contained by some sequences, and contains some sequences: {len(a & b)}")
    print(f"only contains other sequences: {len(newset)}")

    collaspe_dict = {}

    for read in reads_dict:
        if read not in newset:
            continue
        subs = []
        pirs = reads_dict[read].copy()
        for p in pirs:
            subs.append(p)
            if p in reads_dict:
                subs.extend(reads_dict[p])
        subs = set(subs)
        collaspe_dict[read] = list(subs)

    total = []
    with open(os.path.join(outdir, "collapse_info.tsv"), "w") as cof:
        for key, value in collaspe_dict.items():
            total.append(key)
            total.extend(value)
            cof.write(f"{key}\t{len(value)}\t{','.join(value)}\n")

    col_set = set(total)

    print(f"total overlaped reads: {len(col_set)}")

    fa = pysam.FastxFile(originfa)
    with open(os.path.join(outdir, "collapse.fa"), "w") as foa, open(os.path.join(outdir, "uniq.fa"), "w") as uoa:
        for entry in fa:
            if entry.name not in col_set:
                foa.write(f">{entry.name}\n{entry.sequence}\n")
                uoa.write(f">{entry.name}\n{entry.sequence}\n")
            elif entry.name in newset:
                foa.write(f">{entry.name}\n{entry.sequence}\n")

    rna_fetch = pysam.FastaFile(originfa)
    with open(path) as reads, open(os.path.join(outdir, "isoform_info.tsv"), "w") as isof:
        for line in reads:
            fileds = line.split()
            pir1 = fileds[0]
            pir2 = fileds[1]
            pir1s = rna_fetch.fetch(pir1)
            pir2s = rna_fetch.fetch(pir2)
            short = len(pir2s) - len(pir1s)
            pos = pir2s.find(pir1s)
            assert pos >= 0

            if pos == 0:
                info = "3isoform"
            else:
                if pos + len(pir1s) == len(pir2s):
                    info = "5isoform"
                else:
                    info = "53isoform"
            isof.write(f"{pir1}\t{pir2}\t{info}\t{short}\t{pir1s}\t{pir2s}\n")

    with open(path) as reads, open(os.path.join(outdir, f"more_{more_x}_info.tsv"), "w") as c:
        for line in reads:
            info = ""
            fileds = line.split()
            pir1 = fileds[0]
            pir2 = fileds[1]
            pir1s = rna_fetch.fetch(pir1)
            pir2s = rna_fetch.fetch(pir2)
            if len(pir2s) - len(pir1s) > 2:
                continue
            if more_x + pir1s == pir2s:
                info = "L" + more_x
            if pir1s + more_x == pir2s:
                info = "R" + more_x
            if more_x + pir1s + more_x == pir2s:
                info = "LR" + more_x
            if info:
                c.write(f"{pir1}\t{pir2}\t{info}\t{pir1s}\t{pir2s}\n")


def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("-f", dest="originfa", type=str, required=True, help="origin fasta.")
    parser.add_argument("-o", dest="outdir", type=str, required=True, help="output dir.")
    parser.add_argument("-r", dest="reads_overlap", type=str, required=True, help="overlaped reads file, create by bowtie and shell scripts.")
    parser.add_argument("--more_x", dest="more_x", type=str, default="C", choices=["A", "T", "G", "C"], help="more_x analysis.")
    args = parser.parse_args()
    main(args.originfa, args.reads_overlap, args.outdir, args.reads_overlap)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import List, Iterable, Tuple, Iterator

from pybiotk.bx.bitset import BinnedBitSet, MAX


class MergedIntervals:
    def __init__(self, intervals: Iterable[Tuple[int, int]] = []):
        self.bitset = None
        self.intervals = intervals
        self.filled = False

    def _fill(self):
        self.bitset = BinnedBitSet(MAX)
        for interval in self.intervals:
            start, end = int(interval[0]), int(interval[1])
            self.bitset.set_range(start, end - start)
        self.filled = True

    def __iter__(self) -> Iterator[Tuple[int, int]]:
        if not self.filled:
            self._fill()
        end = 0
        while True:
            start = self.bitset.next_set(end)
            if start == self.bitset.size:
                break
            end = self.bitset.next_clear(start)
            yield start, end

    def clear(self):
        del self.bitset
        self.bitset = None

    def extend_intervals(self, intervals: Iterable[Tuple[int, int]]):
        self.intervals.extend(intervals)

    def add_interval(self, interval: Tuple[int, int]):
        self.intervals.append(interval)

    def intersect(self, other: MergedIntervals):
        if not self.filled:
            self._fill()
        if not other.filled:
            other._fill()
        self.bitset.iand(other.bitset)

    def subtract(self, other: MergedIntervals):
        if not self.filled:
            self._fill()
        if not other.filled:
            other._fill()
        other.bitset.invert()
        self.bitset.iand(other.bitset)


def merge_intervals(intervals: Iterable[Tuple[int, int]]) -> List[Tuple[int, int]]:
    intervals_iter = MergedIntervals(intervals)
    interval_list = list(intervals_iter)
    intervals_iter.clear()
    return interval_list


def intersect_intervals(intervals1: Iterable[Tuple[int, int]], intervals2: Iterable[Tuple[int, int]]) -> List[Tuple[int, int]]:
    merged_intervals1 = MergedIntervals(intervals1)
    merged_intervals2 = MergedIntervals(intervals2)
    merged_intervals1.intersect(merged_intervals2)
    merged_intervals2.clear()
    interval_list = list(merged_intervals1)
    merged_intervals1.clear()
    return interval_list


def subtract_intervals(intervals1: Iterable[Tuple[int, int]], intervals2: Iterable[Tuple[int, int]]) -> List[Tuple[int, int]]:
    merged_intervals1 = MergedIntervals(intervals1)
    merged_intervals2 = MergedIntervals(intervals2)
    merged_intervals1.subtract(merged_intervals2)
    interval_list = list(merged_intervals1)
    merged_intervals1.clear()
    return interval_list



#!/usr/bin/env python3
import argparse

from pybiotk.io import FastqFile
from pybiotk.utils import ignore
from stream import stdout


def main(fq_list):
    for fq in fq_list:
        with FastqFile(fq) as f:
            f.to_fasta() | stdout


@ignore
def run():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("input", type=str, nargs="+",
                        help="Input fastq files.")
    args = parser.parse_args()
    main(args.input)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
import warnings
from typing import List, Tuple, Literal, Iterable, Optional

from pybiotk.annodb.anno import GFeature, GenomicAnnotation
from pybiotk.annodb.transcript import Transcript
from pybiotk.intervals import merge_intervals
from stream import flatten, window, skip_while, to_list


class Gene(GFeature):
    def __init__(
        self,
        gene_id: Optional[str] = None,
        gene_name: Optional[str] = None,
        gene_type: Optional[str] = None,
        chrom: Optional[str] = None,
        start: int = 0,
        end: int = 0,
        strand: Optional[Literal['+', '-']] = None,
        cds_start: Optional[int] = None,
        cds_end: Optional[int] = None,
        exons: List[Tuple[int, int]] = None,
        transcripts: List[Transcript] = None,
    ):
        self.gene_id = gene_id
        self.gene_name = gene_name
        self.gene_type = gene_type
        self.chrom = chrom
        self.start = int(start)
        self.end = int(end)
        self.strand = strand
        self.cds_start = int(cds_start) if cds_start is not None else cds_start
        self.cds_end = int(cds_end) if cds_end is not None else cds_end
        self._exons = exons if exons is not None else []
        self.transcripts = transcripts if transcripts is not None else []
        self._introns = None
        self._cds_exons = None
        self._utr5_exons = None
        self._utr3_exons = None

    def __repr__(self) -> str:
        return f"{self.gene_id}:{self.chrom}:{self.start}-{self.end}({self.strand})"

    __str__ = __repr__

    def add_transcript(self, transcript: Transcript):
        if not hasattr(self, "transcripts"):
            self.transcripts = []
        else:
            self.transcripts = list(self.transcripts)
        self.transcripts.append(transcript)

    def post_init(self):
        if self.transcripts:
            min_start = float("inf")
            max_end = 0
            min_cds_start = float("inf")
            max_cds_end = 0
            for transcript in self.transcripts:
                if self.gene_id is None:
                    self.gene_id = transcript.gene_id
                    self.gene_name = transcript.gene_name
                    self.gene_type = transcript.gene_type
                    self.chrom = transcript.chrom
                    self.strand = transcript.strand
                self._exons.extend(transcript.exons())
                if min_start > transcript.start:
                    min_start = transcript.start
                if max_end < transcript.end:
                    max_end = transcript.end
                if transcript.cds_start is None and transcript.cds_end is None:
                    continue
                if min_cds_start > transcript.cds_start:
                    min_cds_start = transcript.cds_start
                if max_cds_end < transcript.cds_end:
                    max_cds_end = transcript.cds_end

            self._exons = list(set(self._exons))
            del self.transcripts
            self.transcripts = []
            self.start = min_start
            self.end = max_end
            if min_cds_start == float("inf"):
                min_cds_start = None
            if max_cds_end == 0:
                max_cds_end = None
            self.cds_start = min_cds_start
            self.cds_end = max_cds_end
        else:
            warnings.warn(
                f"gene({self.gene_id})'s transcripts list is empty, nothing will be done. ")

    @classmethod
    def init_by_transcripts(cls, transcripts: Iterable[Transcript]):
        min_start = float("inf")
        max_end = 0
        min_cds_start = float("inf")
        max_cds_end = 0
        gene_id = None
        gene_name = None
        gene_type = None
        chrom = None
        strand = None
        exons = []
        for transcript in transcripts:
            if gene_id is None:
                gene_id = transcript.gene_id
                gene_name = transcript.gene_name
                gene_type = transcript.gene_type
                chrom = transcript.chrom
                strand = transcript.strand
            exons.extend(transcript.exons())
            if min_start > transcript.start:
                min_start = transcript.start
            if max_end < transcript.end:
                max_end = transcript.end
            if transcript.cds_start is None and transcript.cds_end is None:
                continue
            if min_cds_start > transcript.cds_start:
                min_cds_start = transcript.cds_start
            if max_cds_end < transcript.cds_end:
                max_cds_end = transcript.cds_end
        exons = list(set(exons))
        start = min_start
        end = max_end
        if min_cds_start == float("inf"):
            min_cds_start = None
        if max_cds_end == 0:
            max_cds_end = None
        cds_start = min_cds_start
        cds_end = max_cds_end
        return cls(gene_id, gene_name, gene_type, chrom, start, end, strand, cds_start, cds_end, exons)

    def _classify_exons(self):
        if self.cds_start is not None and self.cds_end is not None:
            utr5_exons = []
            utr3_exons = []
            cds_exons = []
            for exon in self.exons():
                if exon[1] <= self.cds_start:
                    utr5_exons.append(exon)
                elif exon[0] >= self.cds_end:
                    utr3_exons.append(exon)
                elif self.cds_start <= exon[0] and exon[1] <= self.cds_end:
                    cds_exons.append(exon)
                elif exon[0] < self.cds_start < exon[1] <= self.cds_end:
                    utr5_exons.append((exon[0], self.cds_start))
                    cds_exons.append((self.cds_start, exon[1]))
                elif self.cds_start <= exon[0] < self.cds_end < exon[1]:
                    cds_exons.append((exon[0], self.cds_end))
                    utr3_exons.append((self.cds_end, exon[1]))
                elif exon[0] < self.cds_start and self.cds_end < exon[1]:
                    utr5_exons.append((exon[0], self.cds_start))
                    cds_exons.append((self.cds_start, self.cds_end))
                    utr3_exons.append((self.cds_end, exon[1]))

            if self.strand == '-':
                utr5_exons, utr3_exons = utr3_exons, utr5_exons
                self._utr5_exons = utr5_exons
                self._cds_exons = cds_exons
                self._utr3_exons = utr3_exons
            loginfo = f"merged exons: {self.exons()}\ncds_start: {self.cds_start}, cds_end: {self.cds_end}"
            if not cds_exons:
                warnings.warn(f"cds_exons of {self.gene_id} does not exist, please check:\n{loginfo}")
        else:
            self._utr5_exons = []
            self._cds_exons = []
            self._utr3_exons = []

    def is_protein_coding(self) -> bool:
        if self.gene_type == 'protein_coding':
            return True
        else:
            return False

    def exons(self) -> List[Tuple[int, int]]:
        if not self._exons:
            self.post_init()
        self._exons = merge_intervals(self._exons)
        return self._exons

    def introns(self) -> List[Tuple[int, int]]:
        if self._introns is None:
            self._introns = [self.start, *(self.exons() | flatten), self.end] | window(2, 2) | skip_while(lambda x: x[0] == x[1]) | to_list
        return self._introns

    def tss_region(self, region: Tuple[int, int] = (-1000, 1000)) -> Tuple[int, ...]:
        if self.strand == '+':
            region = tuple(i+self.start for i in region)
        else:
            region = tuple(reversed(tuple(self.end-i for i in region)))
        return region

    def downstream(self, down: int = 5000) -> Tuple[int, int]:
        if self.strand == '+':
            downstream = (self.end, self.end+down)
        else:
            downstream = (self.start-down, self.start)
        return downstream

    def cds_exons(self) -> List[Tuple[int, int]]:
        if self._cds_exons is None:
            self._classify_exons()
        return self._cds_exons

    def utr5_exons(self) -> List[Tuple[int, int]]:
        if self._utr5_exons is None:
            self._classify_exons()
        return self._utr5_exons

    def utr3_exons(self) -> List[Tuple[int, int]]:
        if self._utr3_exons is None:
            self._classify_exons()
        return self._utr3_exons

    def annotation(self, blocks: List[Tuple[int, int]], tss_region: Tuple[int, int] = (-1000, 1000), downstream: int = 3000) -> GenomicAnnotation:
        return GenomicAnnotation(self.gene_id, self.gene_name, self.start, self.end, self.gene_type, self.anno(blocks, tss_region, downstream))
